@@ -16,6 +16,12 @@ from rpython.rtyper import rclass
 
 from rpython.rlib.clibffi import FFI_DEFAULT_ABI
 from rpython.rlib.rarithmetic import ovfcheck, r_uint, r_ulonglong
+from rpython.rlib.objectmodel import Symbolic
+
+class LLAsmInfo(object):
+    def __init__(self, lltrace):
+        self.ops_offset = None
+        self.lltrace = lltrace
 
 class LLTrace(object):
     has_been_freed = False
@@ -25,18 +31,27 @@ class LLTrace(object):
         # We need to clone the list of operations because the
         # front-end will mutate them under our feet again.  We also
         # need to make sure things get freed.
-        def mapping(box, _cache={}):
+        _cache={}
+        
+        def mapping(box):
             if isinstance(box, Const) or box is None:
                 return box
             try:
                 newbox = _cache[box]
             except KeyError:
                 newbox = _cache[box] = box.__class__()
+            if hasattr(box, 'accum') and box.accum:
+                newbox.accum = box.accum
             return newbox
         #
         self.inputargs = map(mapping, inputargs)
         self.operations = []
         for op in operations:
+            opnum = op.getopnum()
+            if opnum == rop.GUARD_VALUE:
+                # we don't care about the value 13 here, because we gonna
+                # fish it from the extra slot on frame anyway
+                op.getdescr().make_a_counter_per_value(op, 13)
             if op.getdescr() is not None:
                 if op.is_guard() or op.getopnum() == rop.FINISH:
                     newdescr = op.getdescr()
@@ -46,8 +61,8 @@ class LLTrace(object):
                 newdescr = None
             newop = op.copy_and_change(op.getopnum(),
                                        map(mapping, op.getarglist()),
-                                       mapping(op.result),
                                        newdescr)
+            _cache[op] = newop
             if op.getfailargs() is not None:
                 newop.setfailargs(map(mapping, op.getfailargs()))
             self.operations.append(newop)
@@ -86,15 +101,45 @@ class CallDescr(AbstractDescr):
     def get_result_type(self):
         return getkind(self.RESULT)[0]
 
+    get_normalized_result_type = get_result_type
+
+class TypeIDSymbolic(Symbolic):
+    def __init__(self, STRUCT_OR_ARRAY):
+        self.STRUCT_OR_ARRAY = STRUCT_OR_ARRAY
+
+    def __eq__(self, other):
+        return self.STRUCT_OR_ARRAY is other.STRUCT_OR_ARRAY
+
+    def __ne__(self, other):
+        return not self == other
+
 class SizeDescr(AbstractDescr):
-    def __init__(self, S):
+    def __init__(self, S, vtable, runner):
+        assert not isinstance(vtable, bool)
         self.S = S
+        self._vtable = vtable
+        self._is_object = bool(vtable)
+        self._runner = runner
 
-    def as_vtable_size_descr(self):
-        return self
+    def get_all_fielddescrs(self):
+        return self.all_fielddescrs
 
-    def count_fields_if_immutable(self):
-        return heaptracker.count_fields_if_immutable(self.S)
+    def is_object(self):
+        return self._is_object
+
+    def get_vtable(self):
+        assert self._vtable is not None
+        if self._vtable is Ellipsis:
+            self._vtable = heaptracker.get_vtable_for_gcstruct(self._runner,
+                                                               self.S)
+        return heaptracker.adr2int(llmemory.cast_ptr_to_adr(self._vtable))
+
+    def is_immutable(self):
+        return heaptracker.is_immutable_struct(self.S)
+
+    def get_type_id(self):
+        assert isinstance(self.S, lltype.GcStruct)
+        return TypeIDSymbolic(self.S)     # integer-like symbolic
 
     def __repr__(self):
         return 'SizeDescr(%r)' % (self.S,)
@@ -104,10 +149,21 @@ class FieldDescr(AbstractDescr):
         self.S = S
         self.fieldname = fieldname
         self.FIELD = getattr(S, fieldname)
+        self.index = heaptracker.get_fielddescr_index_in(S, fieldname)
+        self._is_pure = S._immutable_field(fieldname)
+
+    def is_always_pure(self):
+        return self._is_pure
+
+    def get_parent_descr(self):
+        return self.parent_descr
 
     def get_vinfo(self):
         return self.vinfo
 
+    def get_index(self):
+        return self.index
+
     def __repr__(self):
         return 'FieldDescr(%r, %r)' % (self.S, self.fieldname)
 
@@ -146,11 +202,26 @@ def _is_signed_kind(TYPE):
             rffi.cast(TYPE, -1) == -1)
 
 class ArrayDescr(AbstractDescr):
-    def __init__(self, A):
+    all_interiorfielddescrs = None
+    
+    def __init__(self, A, runner):
         self.A = self.OUTERA = A
+        self._is_pure = A._immutable_field(None)
+        self.concrete_type = '\x00'
         if isinstance(A, lltype.Struct):
             self.A = A._flds[A._arrayfld]
 
+    def is_array_of_primitives(self):
+        kind = getkind(self.A.OF)
+        return kind == 'float' or \
+               kind == 'int'
+
+    def is_always_pure(self):
+        return self._is_pure
+
+    def get_all_fielddescrs(self):
+        return self.all_interiorfielddescrs
+
     def __repr__(self):
         return 'ArrayDescr(%r)' % (self.OUTERA,)
 
@@ -170,6 +241,9 @@ class ArrayDescr(AbstractDescr):
         return getkind(self.A.OF) == 'int' \
             and rffi.sizeof(self.A.OF) < symbolic.WORD
 
+    def get_item_size_in_bytes(self):
+        return rffi.sizeof(self.A.OF)
+
     def get_item_integer_min(self):
         if getkind(self.A.OF) != 'int':
             assert False
@@ -184,12 +258,27 @@ class ArrayDescr(AbstractDescr):
         return intbounds.get_integer_max(
             not _is_signed_kind(self.A.OF), rffi.sizeof(self.A.OF))
 
+    def get_type_id(self):
+        assert isinstance(self.A, lltype.GcArray)
+        return TypeIDSymbolic(self.A)     # integer-like symbolic
+
 
 class InteriorFieldDescr(AbstractDescr):
-    def __init__(self, A, fieldname):
+    def __init__(self, A, fieldname, runner):
         self.A = A
         self.fieldname = fieldname
         self.FIELD = getattr(A.OF, fieldname)
+        self.arraydescr = runner.arraydescrof(A)
+        self.fielddescr = runner.fielddescrof(A.OF, fieldname)
+
+    def get_index(self):
+        return self.fielddescr.get_index()
+
+    def get_arraydescr(self):
+        return self.arraydescr
+
+    def get_field_descr(self):
+        return self.fielddescr
 
     def __repr__(self):
         return 'InteriorFieldDescr(%r, %r)' % (self.A, self.fieldname)
@@ -232,8 +321,13 @@ class LLGraphCPU(model.AbstractCPU):
     supports_floats = True
     supports_longlong = r_uint is not r_ulonglong
     supports_singlefloats = True
+    supports_guard_gc_type = True
     translate_support_code = False
     is_llgraph = True
+    vector_extension = True
+    vector_register_size = 16 # in bytes
+    vector_horizontal_operations = True
+    vector_pack_slots = True
 
     def __init__(self, rtyper, stats=None, *ignored_args, **kwds):
         model.AbstractCPU.__init__(self)
@@ -245,6 +339,9 @@ class LLGraphCPU(model.AbstractCPU):
         self.stats = stats or MiniStats()
         self.vinfo_for_tests = kwds.get('vinfo_for_tests', None)
 
+    def stitch_bridge(self, faildescr, target):
+        faildescr._llgraph_bridge = target[0].lltrace
+
     def compile_loop(self, inputargs, operations, looptoken, jd_id=0,
                      unique_id=0, log=True, name='', logger=None):
         clt = model.CompiledLoopToken(self, looptoken.number)
@@ -262,6 +359,7 @@ class LLGraphCPU(model.AbstractCPU):
         faildescr._llgraph_bridge = lltrace
         clt._llgraph_alltraces.append(lltrace)
         self._record_labels(lltrace)
+        return LLAsmInfo(lltrace)
 
     def _record_labels(self, lltrace):
         for i, op in enumerate(lltrace.operations):
@@ -273,16 +371,16 @@ class LLGraphCPU(model.AbstractCPU):
             trace.invalid = True
 
     def redirect_call_assembler(self, oldlooptoken, newlooptoken):
-        oldtrace = oldlooptoken.compiled_loop_token._llgraph_loop
-        newtrace = newlooptoken.compiled_loop_token._llgraph_loop
+        oldc = oldlooptoken.compiled_loop_token
+        newc = newlooptoken.compiled_loop_token
+        oldtrace = oldc._llgraph_loop
+        newtrace = newc._llgraph_loop
         OLD = [box.type for box in oldtrace.inputargs]
         NEW = [box.type for box in newtrace.inputargs]
         assert OLD == NEW
-        assert not hasattr(oldlooptoken, '_llgraph_redirected')
-        oldlooptoken.compiled_loop_token._llgraph_redirected = True
-        oldlooptoken.compiled_loop_token._llgraph_loop = newtrace
-        alltraces = newlooptoken.compiled_loop_token._llgraph_alltraces
-        oldlooptoken.compiled_loop_token._llgraph_alltraces = alltraces
+        assert not hasattr(oldc, '_llgraph_redirected')
+        oldc._llgraph_redirected = newc
+        oldc._llgraph_alltraces = newc._llgraph_alltraces
 
     def free_loop_and_bridges(self, compiled_loop_token):
         for c in compiled_loop_token._llgraph_alltraces:
@@ -295,7 +393,10 @@ class LLGraphCPU(model.AbstractCPU):
         return self._execute_token
 
     def _execute_token(self, loop_token, *args):
-        lltrace = loop_token.compiled_loop_token._llgraph_loop
+        loopc = loop_token.compiled_loop_token
+        while hasattr(loopc, '_llgraph_redirected'):
+            loopc = loopc._llgraph_redirected
+        lltrace = loopc._llgraph_loop
         frame = LLFrame(self, lltrace.inputargs, args)
         try:
             frame.execute(lltrace)
@@ -303,6 +404,18 @@ class LLGraphCPU(model.AbstractCPU):
         except ExecutionFinished, e:
             return e.deadframe
 
+    def get_value_direct(self, deadframe, tp, index):
+        v = deadframe._extra_value
+        if tp == 'i':
+            assert lltype.typeOf(v) == lltype.Signed
+        elif tp == 'r':
+            assert lltype.typeOf(v) == llmemory.GCREF
+        elif tp == 'f':
+            assert lltype.typeOf(v) == longlong.FLOATSTORAGE
+        else:
+            assert False
+        return v
+
     def get_int_value(self, deadframe, index):
         v = deadframe._values[index]
         assert lltype.typeOf(v) == lltype.Signed
@@ -336,10 +449,10 @@ class LLGraphCPU(model.AbstractCPU):
         values = []
         for box in frame.force_guard_op.getfailargs():
             if box is not None:
-                if box is not frame.current_op.result:
+                if box is not frame.current_op:
                     value = frame.env[box]
                 else:
-                    value = box.value    # 0 or 0.0 or NULL
+                    value = box.getvalue()    # 0 or 0.0 or NULL
             else:
                 value = None
             values.append(value)
@@ -367,14 +480,19 @@ class LLGraphCPU(model.AbstractCPU):
             self.descrs[key] = descr
             return descr
 
-    def sizeof(self, S):
+    def sizeof(self, S, vtable=lltype.nullptr(rclass.OBJECT_VTABLE)):
         key = ('size', S)
         try:
-            return self.descrs[key]
+            descr = self.descrs[key]
         except KeyError:
-            descr = SizeDescr(S)
+            descr = SizeDescr(S, vtable, self)
             self.descrs[key] = descr
-            return descr
+            descr.all_fielddescrs = heaptracker.all_fielddescrs(self, S,
+                    get_field_descr=LLGraphCPU.fielddescrof)
+        if descr._is_object and vtable is not Ellipsis:
+            assert vtable
+            heaptracker.testing_gcstruct2vtable.setdefault(S, vtable)
+        return descr
 
     def fielddescrof(self, S, fieldname):
         key = ('field', S, fieldname)
@@ -383,6 +501,12 @@ class LLGraphCPU(model.AbstractCPU):
         except KeyError:
             descr = FieldDescr(S, fieldname)
             self.descrs[key] = descr
+            if (isinstance(S, lltype.GcStruct) and
+                    heaptracker.has_gcstruct_a_vtable(S)):
+                vtable = Ellipsis
+            else:
+                vtable = None
+            descr.parent_descr = self.sizeof(S, vtable)
             if self.vinfo_for_tests is not None:
                 descr.vinfo = self.vinfo_for_tests
             return descr
@@ -392,8 +516,12 @@ class LLGraphCPU(model.AbstractCPU):
         try:
             return self.descrs[key]
         except KeyError:
-            descr = ArrayDescr(A)
+            descr = ArrayDescr(A, self)
             self.descrs[key] = descr
+            if isinstance(A, lltype.Array) and isinstance(A.OF, lltype.Struct):
+                descrs = heaptracker.all_interiorfielddescrs(self,
+                        A, get_field_descr=LLGraphCPU.interiorfielddescrof)
+                descr.all_interiorfielddescrs = descrs
             return descr
 
     def interiorfielddescrof(self, A, fieldname):
@@ -401,7 +529,7 @@ class LLGraphCPU(model.AbstractCPU):
         try:
             return self.descrs[key]
         except KeyError:
-            descr = InteriorFieldDescr(A, fieldname)
+            descr = InteriorFieldDescr(A, fieldname, self)
             self.descrs[key] = descr
             return descr
 
@@ -435,6 +563,22 @@ class LLGraphCPU(model.AbstractCPU):
             self.descrs[key] = descr
             return descr
 
+    def check_is_object(self, gcptr):
+        """Check if the given, non-null gcptr refers to an rclass.OBJECT
+        or not at all (an unrelated GcStruct or a GcArray).  Only usable
+        in the llgraph backend, or after translation of a real backend."""
+        ptr = lltype.normalizeptr(gcptr._obj.container._as_ptr())
+        T = lltype.typeOf(ptr).TO
+        return heaptracker.has_gcstruct_a_vtable(T) or T is rclass.OBJECT
+
+    def get_actual_typeid(self, gcptr):
+        """Fetch the actual typeid of the given gcptr, as an integer.
+        Only usable in the llgraph backend, or after translation of a
+        real backend.  (Here in the llgraph backend, returns a
+        TypeIDSymbolic instead of a real integer.)"""
+        ptr = lltype.normalizeptr(gcptr._obj.container._as_ptr())
+        return TypeIDSymbolic(lltype.typeOf(ptr).TO)
+
     # ------------------------------------------------------------
 
     def maybe_on_top_of_llinterp(self, func, args, RESULT):
@@ -461,13 +605,17 @@ class LLGraphCPU(model.AbstractCPU):
         p = support.cast_arg(lltype.Ptr(descr.S), p)
         return support.cast_result(descr.FIELD, getattr(p, descr.fieldname))
 
-    bh_getfield_gc_pure = bh_getfield_gc
+    bh_getfield_gc_pure_i = bh_getfield_gc
+    bh_getfield_gc_pure_r = bh_getfield_gc
+    bh_getfield_gc_pure_f = bh_getfield_gc
     bh_getfield_gc_i = bh_getfield_gc
     bh_getfield_gc_r = bh_getfield_gc
     bh_getfield_gc_f = bh_getfield_gc
 
     bh_getfield_raw = bh_getfield_gc
-    bh_getfield_raw_pure = bh_getfield_raw
+    bh_getfield_raw_pure_i = bh_getfield_raw
+    bh_getfield_raw_pure_r = bh_getfield_raw
+    bh_getfield_raw_pure_f = bh_getfield_raw
     bh_getfield_raw_i = bh_getfield_raw
     bh_getfield_raw_r = bh_getfield_raw
     bh_getfield_raw_f = bh_getfield_raw
@@ -495,13 +643,17 @@ class LLGraphCPU(model.AbstractCPU):
         array = a._obj
         return support.cast_result(descr.A.OF, array.getitem(index))
 
-    bh_getarrayitem_gc_pure = bh_getarrayitem_gc
+    bh_getarrayitem_gc_pure_i = bh_getarrayitem_gc
+    bh_getarrayitem_gc_pure_r = bh_getarrayitem_gc
+    bh_getarrayitem_gc_pure_f = bh_getarrayitem_gc
     bh_getarrayitem_gc_i = bh_getarrayitem_gc
     bh_getarrayitem_gc_r = bh_getarrayitem_gc
     bh_getarrayitem_gc_f = bh_getarrayitem_gc
 
     bh_getarrayitem_raw = bh_getarrayitem_gc
-    bh_getarrayitem_raw_pure = bh_getarrayitem_raw
+    bh_getarrayitem_raw_pure_i = bh_getarrayitem_raw
+    bh_getarrayitem_raw_pure_r = bh_getarrayitem_raw
+    bh_getarrayitem_raw_pure_f = bh_getarrayitem_raw
     bh_getarrayitem_raw_i = bh_getarrayitem_raw
     bh_getarrayitem_raw_r = bh_getarrayitem_raw
     bh_getarrayitem_raw_f = bh_getarrayitem_raw
@@ -634,11 +786,11 @@ class LLGraphCPU(model.AbstractCPU):
         return lltype.cast_opaque_ptr(llmemory.GCREF,
                                       lltype.malloc(sizedescr.S, zero=True))
 
-    def bh_new_with_vtable(self, vtable, descr):
+    def bh_new_with_vtable(self, descr):
         result = lltype.malloc(descr.S, zero=True)
         result_as_objptr = lltype.cast_pointer(rclass.OBJECTPTR, result)
         result_as_objptr.typeptr = support.cast_from_int(rclass.CLASSTYPE,
-                                                         vtable)
+                                                descr.get_vtable())
         return lltype.cast_opaque_ptr(llmemory.GCREF, result)
 
     def bh_new_array(self, length, arraydescr):
@@ -658,6 +810,145 @@ class LLGraphCPU(model.AbstractCPU):
     def bh_new_raw_buffer(self, size):
         return lltype.malloc(rffi.CCHARP.TO, size, flavor='raw')
 
+    # vector operations
+    vector_arith_code = """
+    def bh_vec_{0}_{1}(self, vx, vy, count):
+        assert len(vx) == len(vy) == count
+        return [_vx {2} _vy for _vx,_vy in zip(vx,vy)]
+    """
+    exec py.code.Source(vector_arith_code.format('int','add','+')).compile()
+    exec py.code.Source(vector_arith_code.format('int','sub','-')).compile()
+    exec py.code.Source(vector_arith_code.format('int','mul','*')).compile()
+    exec py.code.Source(vector_arith_code.format('int','and','&')).compile()
+    exec py.code.Source(vector_arith_code.format('int','or','|')).compile()
+    exec py.code.Source(vector_arith_code.format('float','add','+')).compile()
+    exec py.code.Source(vector_arith_code.format('float','sub','-')).compile()
+    exec py.code.Source(vector_arith_code.format('float','mul','*')).compile()
+    exec py.code.Source(vector_arith_code.format('float','truediv','/')).compile()
+    exec py.code.Source(vector_arith_code.format('float','eq','==')).compile()
+
+    def bh_vec_float_neg(self, vx, count):
+        return [e * -1 for e in vx]
+
+    def bh_vec_float_abs(self, vx, count):
+        return [abs(e) for e in vx]
+
+    def bh_vec_float_eq(self, vx, vy, count):
+        assert len(vx) == len(vy) == count
+        return [_vx == _vy for _vx,_vy in zip(vx,vy)]
+
+    def bh_vec_float_ne(self, vx, vy, count):
+        assert len(vx) == len(vy) == count
+        return [_vx != _vy for _vx,_vy in zip(vx,vy)]
+
+    bh_vec_int_eq = bh_vec_float_eq
+    bh_vec_int_ne = bh_vec_float_ne
+
+    def bh_vec_int_is_true(self, vx, count):
+        return map(lambda x: bool(x), vx)
+
+    def bh_vec_int_is_false(self, vx, count):
+        return map(lambda x: not bool(x), vx)
+
+    def bh_vec_int_xor(self, vx, vy, count):
+        return [int(x) ^ int(y) for x,y in zip(vx,vy)]
+
+    def bh_vec_cast_float_to_singlefloat(self, vx, count):
+        from rpython.rlib.rarithmetic import r_singlefloat
+        return [longlong.singlefloat2int(r_singlefloat(longlong.getrealfloat(v)))
+                for v in vx]
+
+    def bh_vec_cast_singlefloat_to_float(self, vx, count):
+        return [longlong.getfloatstorage(float(longlong.int2singlefloat(v)))
+                for v in vx]
+
+        a = float(a)
+        return longlong.getfloatstorage(a)
+
+    def bh_vec_cast_float_to_int(self, vx, count):
+        return [int(x) for x in vx]
+
+    def bh_vec_cast_int_to_float(self, vx, count):
+        return [float(x) for x in vx]
+
+    def bh_vec_f(self, count):
+        return [0.0] * count
+
+    def bh_vec_i(self, count):
+        return [0] * count
+
+    def _bh_vec_pack(self, tv, sv, index, count, newcount):
+        while len(tv) < newcount: tv.append(None)
+        if not isinstance(sv, list):
+            tv[index] = sv
+            return tv
+        for i in range(count):
+            tv[index+i] = sv[i]
+        return tv
+
+    bh_vec_pack_f = _bh_vec_pack
+    bh_vec_pack_i = _bh_vec_pack
+
+    def _bh_vec_unpack(self, vx, index, count, newcount):
+        return vx[index:index+count]
+
+    bh_vec_unpack_f = _bh_vec_unpack
+    bh_vec_unpack_i = _bh_vec_unpack
+
+    def _bh_vec_expand(self, x, count):
+        return [x] * count
+
+    bh_vec_expand_f = _bh_vec_expand
+    bh_vec_expand_i = _bh_vec_expand
+
+    def bh_vec_int_signext(self, vx, ext, count):
+        return [heaptracker.int_signext(_vx, ext) for _vx in vx]
+
+    def build_getarrayitem(func):
+        def method(self, struct, offset, descr, _count):
+            values = []
+            count = self.vector_register_size // descr.get_item_size_in_bytes()
+            assert _count == count
+            assert count > 0
+            for i in range(count):
+                val = func(self, struct, offset + i, descr)
+                values.append(val)
+            return values
+        return method
+
+    bh_vec_getarrayitem_gc_i = build_getarrayitem(bh_getarrayitem_gc)
+    bh_vec_getarrayitem_gc_f = build_getarrayitem(bh_getarrayitem_gc)
+    bh_vec_getarrayitem_raw_i = build_getarrayitem(bh_getarrayitem_raw)
+    bh_vec_getarrayitem_raw_f = build_getarrayitem(bh_getarrayitem_raw)
+    del build_getarrayitem
+
+    def _bh_vec_raw_load(self, struct, offset, descr, _count):
+        values = []
+        stride = descr.get_item_size_in_bytes()
+        count = self.vector_register_size // descr.get_item_size_in_bytes()
+        assert _count == count
+        assert count > 0
+        for i in range(count):
+            val = self.bh_raw_load(struct, offset + i*stride, descr)
+            values.append(val)
+        return values
+
+    bh_vec_raw_load_i = _bh_vec_raw_load
+    bh_vec_raw_load_f = _bh_vec_raw_load
+
+    def bh_vec_raw_store(self, struct, offset, newvalues, descr, count):
+        stride = descr.get_item_size_in_bytes()
+        for i,n in enumerate(newvalues):
+            self.bh_raw_store(struct, offset + i*stride, n, descr)
+
+    def bh_vec_setarrayitem_raw(self, struct, offset, newvalues, descr, count):
+        for i,n in enumerate(newvalues):
+            self.bh_setarrayitem_raw(struct, offset + i, n, descr)
+
+    def bh_vec_setarrayitem_gc(self, struct, offset, newvalues, descr, count):
+        for i,n in enumerate(newvalues):
+            self.bh_setarrayitem_gc(struct, offset + i, n, descr)
+
     def store_fail_descr(self, deadframe, descr):
         pass # I *think*
 
@@ -667,11 +958,13 @@ class LLDeadFrame(object):
     _TYPE = llmemory.GCREF
 
     def __init__(self, latest_descr, values,
-                 last_exception=None, saved_data=None):
+                 last_exception=None, saved_data=None,
+                 extra_value=None):
         self._latest_descr = latest_descr
         self._values = values
         self._last_exception = last_exception
         self._saved_data = saved_data
+        self._extra_value = extra_value
 
 
 class LLFrame(object):
@@ -706,7 +999,18 @@ class LLFrame(object):
         return hash(self)
 
     def setenv(self, box, arg):
-        if box.type == INT:
+        if box.is_vector() and box.count > 1:
+            if box.datatype == INT:
+                for i,a in enumerate(arg):
+                    if isinstance(a, bool):
+                        arg[i] = int(a) 
+                assert all([lltype.typeOf(a) == lltype.Signed for a in arg])
+            elif box.datatype == FLOAT:
+                assert all([lltype.typeOf(a) == longlong.FLOATSTORAGE or \
+                            lltype.typeOf(a) == lltype.Signed for a in arg])
+            else:
+                raise AssertionError(box)
+        elif box.type == INT:
             # typecheck the result
             if isinstance(arg, bool):
                 arg = int(arg)
@@ -749,8 +1053,8 @@ class LLFrame(object):
                     i = 0
                 self.do_renaming(targetargs, j.args)
                 continue
-            if op.result is not None:
-                self.setenv(op.result, resval)
+            if op.type != 'v':
+                self.setenv(op, resval)
             else:
                 assert resval is None
             i += 1
@@ -764,7 +1068,26 @@ class LLFrame(object):
 
     # -----------------------------------------------------
 
-    def fail_guard(self, descr, saved_data=None):
+    def _accumulate(self, descr, failargs, values):
+        info = descr.rd_vector_info
+        while info:
+            i = info.getpos_in_failargs()
+            value = values[i]
+            assert isinstance(value, list)
+            if info.accum_operation == '+':
+                value = sum(value)
+            elif info.accum_operation == '*':
+                def prod(acc, x): return acc * x
+                value = reduce(prod, value, 1)
+            else:
+                raise NotImplementedError("accum operator in fail guard")
+            values[i] = value
+            info = info.next()
+
+    def fail_guard(self, descr, saved_data=None, extra_value=None,
+                   propagate_exception=False):
+        if not propagate_exception:
+            assert self.last_exception is None
         values = []
         for box in self.current_op.getfailargs():
             if box is not None:
@@ -772,14 +1095,19 @@ class LLFrame(object):
             else:
                 value = None
             values.append(value)
+        self._accumulate(descr, self.current_op.getfailargs(), values)
         if hasattr(descr, '_llgraph_bridge'):
+            if propagate_exception:
+                assert (descr._llgraph_bridge.operations[0].opnum in
+                        (rop.SAVE_EXC_CLASS, rop.GUARD_EXCEPTION,
+                         rop.GUARD_NO_EXCEPTION))
             target = (descr._llgraph_bridge, -1)
             values = [value for value in values if value is not None]
             raise Jump(target, values)
         else:
             raise ExecutionFinished(LLDeadFrame(descr, values,
                                                 self.last_exception,
-                                                saved_data))
+                                                saved_data, extra_value))
 
     def execute_force_spill(self, _, arg):
         pass
@@ -791,17 +1119,27 @@ class LLFrame(object):
         argboxes = self.current_op.getarglist()
         self.do_renaming(argboxes, args)
 
+    def _test_true(self, arg):
+        if isinstance(arg, list):
+            return all(arg)
+        return arg
+
+    def _test_false(self, arg):
+        if isinstance(arg, list):
+            return any(arg)
+        return arg
+
     def execute_guard_true(self, descr, arg):
-        if not arg:
+        if not self._test_true(arg):
             self.fail_guard(descr)
 
     def execute_guard_false(self, descr, arg):
-        if arg:
+        if self._test_false(arg):
             self.fail_guard(descr)
 
     def execute_guard_value(self, descr, arg1, arg2):
         if arg1 != arg2:
-            self.fail_guard(descr)
+            self.fail_guard(descr, extra_value=arg1)
 
     def execute_guard_nonnull(self, descr, arg):
         if not arg:
@@ -823,9 +1161,35 @@ class LLFrame(object):
         self.execute_guard_nonnull(descr, arg)
         self.execute_guard_class(descr, arg, klass)
 
+    def execute_guard_gc_type(self, descr, arg, typeid):
+        assert isinstance(typeid, TypeIDSymbolic)
+        TYPE = arg._obj.container._TYPE
+        if TYPE != typeid.STRUCT_OR_ARRAY:
+            self.fail_guard(descr)
+
+    def execute_guard_is_object(self, descr, arg):
+        TYPE = arg._obj.container._TYPE
+        while TYPE is not rclass.OBJECT:
+            if not isinstance(TYPE, lltype.GcStruct):   # or TYPE is None
+                self.fail_guard(descr)
+                return
+            _, TYPE = TYPE._first_struct()
+
+    def execute_guard_subclass(self, descr, arg, klass):
+        value = lltype.cast_opaque_ptr(rclass.OBJECTPTR, arg)
+        expected_class = llmemory.cast_adr_to_ptr(
+            llmemory.cast_int_to_adr(klass),
+            rclass.CLASSTYPE)
+        if (expected_class.subclassrange_min
+                <= value.typeptr.subclassrange_min
+                <= expected_class.subclassrange_max):
+            pass
+        else:
+            self.fail_guard(descr)
+
     def execute_guard_no_exception(self, descr):
         if self.last_exception is not None:
-            self.fail_guard(descr)
+            self.fail_guard(descr, propagate_exception=True)
 
     def execute_guard_exception(self, descr, excklass):
         lle = self.last_exception
@@ -837,7 +1201,7 @@ class LLFrame(object):
             llmemory.cast_int_to_adr(excklass),
             rclass.CLASSTYPE)
         if gotklass != excklass:
-            self.fail_guard(descr)
+            self.fail_guard(descr, propagate_exception=True)
         #
         res = lle.args[1]
         self.last_exception = None
@@ -846,7 +1210,7 @@ class LLFrame(object):
     def execute_guard_not_forced(self, descr):
         if self.forced_deadframe is not None:
             saved_data = self.forced_deadframe._saved_data
-            self.fail_guard(descr, saved_data)
+            self.fail_guard(descr, saved_data, propagate_exception=True)
         self.force_guard_op = self.current_op
     execute_guard_not_forced_2 = execute_guard_not_forced
 
@@ -908,9 +1272,9 @@ class LLFrame(object):
         if not cond:
             return
         # cond_call can't have a return value
-        self.execute_call(calldescr, func, *args)
+        self.execute_call_n(calldescr, func, *args)
 
-    def execute_call(self, calldescr, func, *args):
+    def _execute_call(self, calldescr, func, *args):
         effectinfo = calldescr.get_extra_info()
         if effectinfo is not None and hasattr(effectinfo, 'oopspecindex'):
             oopspecindex = effectinfo.oopspecindex
@@ -926,22 +1290,31 @@ class LLFrame(object):
             res = _example_res[getkind(TP.RESULT)[0]]
         return res
 
-    def execute_call_may_force(self, calldescr, func, *args):
-        call_op = self.lltrace.operations[self.current_index]
+    execute_call_i = _execute_call
+    execute_call_r = _execute_call
+    execute_call_f = _execute_call
+    execute_call_n = _execute_call
+
+    def _execute_call_may_force(self, calldescr, func, *args):
         guard_op = self.lltrace.operations[self.current_index + 1]
         assert guard_op.getopnum() == rop.GUARD_NOT_FORCED
         self.force_guard_op = guard_op
-        res = self.execute_call(calldescr, func, *args)
+        res = self._execute_call(calldescr, func, *args)
         del self.force_guard_op
         return res
 
-    def execute_call_release_gil(self, descr, saveerr, func, *args):
+    execute_call_may_force_n = _execute_call_may_force
+    execute_call_may_force_r = _execute_call_may_force
+    execute_call_may_force_f = _execute_call_may_force
+    execute_call_may_force_i = _execute_call_may_force
+
+    def _execute_call_release_gil(self, descr, saveerr, func, *args):
         if hasattr(descr, '_original_func_'):
             func = descr._original_func_     # see pyjitpl.py
             # we want to call the function that does the aroundstate
             # manipulation here (as a hack, instead of really doing
             # the aroundstate manipulation ourselves)
-            return self.execute_call_may_force(descr, func, *args)
+            return self._execute_call_may_force(descr, func, *args)
         guard_op = self.lltrace.operations[self.current_index + 1]
         assert guard_op.getopnum() == rop.GUARD_NOT_FORCED
         self.force_guard_op = guard_op
@@ -960,61 +1333,74 @@ class LLFrame(object):
         del self.force_guard_op
         return support.cast_result(descr.RESULT, result)
 
-    def execute_call_assembler(self, descr, *args):
-        # XXX simplify the following a bit
-        #
-        # pframe = CALL_ASSEMBLER(args..., descr=looptoken)
-        # ==>
-        #     pframe = CALL looptoken.loopaddr(*args)
-        #     JUMP_IF_FAST_PATH @fastpath
-        #     res = CALL assembler_call_helper(pframe)
-        #     jmp @done
-        #   @fastpath:
-        #     res = GETFIELD(pframe, 'result')
-        #   @done:
-        #
-        call_op = self.lltrace.operations[self.current_index]
-        guard_op = self.lltrace.operations[self.current_index + 1]
-        assert guard_op.getopnum() == rop.GUARD_NOT_FORCED
-        self.force_guard_op = guard_op
-        pframe = self.cpu._execute_token(descr, *args)
-        del self.force_guard_op
-        #
-        jd = descr.outermost_jitdriver_sd
-        assert jd is not None, ("call_assembler(): the loop_token needs "
-                                "to have 'outermost_jitdriver_sd'")
-        if jd.index_of_virtualizable != -1:
-            vable = args[jd.index_of_virtualizable]
-        else:
-            vable = lltype.nullptr(llmemory.GCREF.TO)
-        #
-        # Emulate the fast path
-        #
-        faildescr = self.cpu.get_latest_descr(pframe)
-        if faildescr == self.cpu.done_with_this_frame_descr_int:
-            return self.cpu.get_int_value(pframe, 0)
-        elif faildescr == self.cpu.done_with_this_frame_descr_ref:
-            return self.cpu.get_ref_value(pframe, 0)
-        elif faildescr == self.cpu.done_with_this_frame_descr_float:
-            return self.cpu.get_float_value(pframe, 0)
-        elif faildescr == self.cpu.done_with_this_frame_descr_void:
-            return None
+    execute_call_release_gil_n = _execute_call_release_gil
+    execute_call_release_gil_i = _execute_call_release_gil
+    execute_call_release_gil_r = _execute_call_release_gil
+    execute_call_release_gil_f = _execute_call_release_gil
 
-        assembler_helper_ptr = jd.assembler_helper_adr.ptr  # fish
-        try:
-            result = assembler_helper_ptr(pframe, vable)
-        except LLException, lle:
-            assert self.last_exception is None, "exception left behind"
-            self.last_exception = lle
-            # fish op
-            op = self.current_op
-            return op.result and op.result.value
-        if isinstance(result, float):
-            result = support.cast_to_floatstorage(result)
-        return result
+    def _new_execute_call_assembler(def_val):
+        def _execute_call_assembler(self, descr, *args):
+            # XXX simplify the following a bit
+            #
+            # pframe = CALL_ASSEMBLER(args..., descr=looptoken)
+            # ==>
+            #     pframe = CALL looptoken.loopaddr(*args)
+            #     JUMP_IF_FAST_PATH @fastpath
+            #     res = CALL assembler_call_helper(pframe)
+            #     jmp @done
+            #   @fastpath:
+            #     res = GETFIELD(pframe, 'result')
+            #   @done:
+            #
+            call_op = self.lltrace.operations[self.current_index]
+            guard_op = self.lltrace.operations[self.current_index + 1]
+            assert guard_op.getopnum() == rop.GUARD_NOT_FORCED
+            self.force_guard_op = guard_op
+            pframe = self.cpu._execute_token(descr, *args)
+            del self.force_guard_op
+            #
+            jd = descr.outermost_jitdriver_sd
+            assert jd is not None, ("call_assembler(): the loop_token needs "
+                                    "to have 'outermost_jitdriver_sd'")
+            if jd.index_of_virtualizable != -1:
+                vable = args[jd.index_of_virtualizable]
+            else:
+                vable = lltype.nullptr(llmemory.GCREF.TO)
+            #
+            # Emulate the fast path
+            #
+            faildescr = self.cpu.get_latest_descr(pframe)
+            if faildescr == self.cpu.done_with_this_frame_descr_int:
+                return self.cpu.get_int_value(pframe, 0)
+            elif faildescr == self.cpu.done_with_this_frame_descr_ref:
+                return self.cpu.get_ref_value(pframe, 0)
+            elif faildescr == self.cpu.done_with_this_frame_descr_float:
+                return self.cpu.get_float_value(pframe, 0)
+            elif faildescr == self.cpu.done_with_this_frame_descr_void:
+                return None
 
-    def execute_same_as(self, _, x):
+            assembler_helper_ptr = jd.assembler_helper_adr.ptr  # fish
+            try:
+                result = assembler_helper_ptr(pframe, vable)
+            except LLException, lle:
+                assert self.last_exception is None, "exception left behind"
+                self.last_exception = lle
+                # fish op
+                result = def_val
+            if isinstance(result, float):
+                result = support.cast_to_floatstorage(result)
+            return result
+        return _execute_call_assembler
+
+    execute_call_assembler_i = _new_execute_call_assembler(0)
+    execute_call_assembler_r = _new_execute_call_assembler(lltype.nullptr(llmemory.GCREF.TO))
+    execute_call_assembler_f = _new_execute_call_assembler(0.0)
+    execute_call_assembler_n = _new_execute_call_assembler(None)
+
+    def execute_same_as_i(self, _, x):
         return x
+    execute_same_as_f = execute_same_as_i
+    execute_same_as_r = execute_same_as_i
 
     def execute_debug_merge_point(self, descr, *args):
         from rpython.jit.metainterp.warmspot import get_stats
@@ -1031,9 +1417,8 @@ class LLFrame(object):
     def execute_leave_portal_frame(self, descr, *args):
         pass
 
-    def execute_new_with_vtable(self, _, vtable):
-        descr = heaptracker.vtable2descr(self.cpu, vtable)
-        return self.cpu.bh_new_with_vtable(vtable, descr)
+    def execute_new_with_vtable(self, descr):
+        return self.cpu.bh_new_with_vtable(descr)
 
     def execute_force_token(self, _):
         return self
@@ -1047,6 +1432,33 @@ class LLFrame(object):
     def execute_keepalive(self, descr, x):
         pass
 
+    def execute_save_exc_class(self, descr):
+        lle = self.last_exception
+        if lle is None:
+            return 0
+        else:
+            return support.cast_to_int(lle.args[0])
+
+    def execute_save_exception(self, descr):
+        lle = self.last_exception
+        if lle is None:
+            res = lltype.nullptr(llmemory.GCREF.TO)
+        else:
+            res = lltype.cast_opaque_ptr(llmemory.GCREF, lle.args[1])
+        self.last_exception = None
+        return res
+
+    def execute_restore_exception(self, descr, kls, e):
+        kls = heaptracker.int2adr(kls)
+        if e:
+            value = lltype.cast_opaque_ptr(rclass.OBJECTPTR, e)
+            assert llmemory.cast_ptr_to_adr(value.typeptr) == kls
+            lle = LLException(value.typeptr, e)
+        else:
+            assert kls == llmemory.NULL
+            lle = None
+        self.last_exception = lle
+
 
 def _getdescr(op):
     d = op.getdescr()
@@ -1080,7 +1492,21 @@ def _setup():
                 new_args = args + (descr,)
             else:
                 new_args = args
-            return getattr(self.cpu, 'bh_' + opname)(*new_args)
+            if opname.startswith('vec_'):
+                # pre vector op
+                count = self.current_op.count
+                assert count >= 1
+                new_args = new_args + (count,)
+            result = getattr(self.cpu, 'bh_' + opname)(*new_args)
+            if isinstance(result, list):
+                # post vector op
+                count = self.current_op.count
+                if len(result) > count:
+                    assert count > 0
+                    result = result[:count]
+                if count == 1:
+                    result = result[0]
+            return result
         execute.func_name = 'execute_' + opname
         return execute
 
