@@ -80,7 +80,7 @@ class ExplicitVirtualizableTests:
             return xy.inst_x
         res = self.meta_interp(f, [20])
         assert res == 30
-        self.check_simple_loop(setfield_gc=0, getfield_gc=0)
+        self.check_simple_loop(setfield_gc=0, getfield_gc_i=0)
 
     def test_preexisting_access_2(self):
         myjitdriver = JitDriver(greens = [], reds = ['n', 'xy'],
@@ -105,8 +105,8 @@ class ExplicitVirtualizableTests:
         assert f(5) == 185
         res = self.meta_interp(f, [5])
         assert res == 185
-        self.check_resops(setfield_gc=0,
-                          getfield_gc=2)  # <= at the header of the loop
+        self.check_resops(setfield_gc=0, getfield_gc_r=1,
+                          getfield_gc_i=1)  # <= at the header of the loop
 
     def test_two_paths_access(self):
         myjitdriver = JitDriver(greens = [], reds = ['n', 'xy'],
@@ -128,7 +128,7 @@ class ExplicitVirtualizableTests:
             return xy.inst_x
         res = self.meta_interp(f, [18])
         assert res == 10118
-        self.check_resops(setfield_gc=0, getfield_gc=2)
+        self.check_resops(setfield_gc=0, getfield_gc_i=1, getfield_gc_r=1)
 
     def test_synchronize_in_return(self):
         myjitdriver = JitDriver(greens = [], reds = ['n', 'xy'],
@@ -152,7 +152,7 @@ class ExplicitVirtualizableTests:
             return xy.inst_x
         res = self.meta_interp(f, [18])
         assert res == 10180
-        self.check_resops(setfield_gc=0, getfield_gc=2)
+        self.check_resops(setfield_gc=0, getfield_gc_i=1, getfield_gc_r=1)
 
     def test_synchronize_in_return_2(self):
         myjitdriver = JitDriver(greens = [], reds = ['n', 'xy'],
@@ -207,7 +207,7 @@ class ExplicitVirtualizableTests:
             return res
         res = self.meta_interp(f, [40])
         assert res == 50 * 4
-        self.check_resops(setfield_gc=0, getfield_gc=4)
+        self.check_resops(setfield_gc=0, getfield_gc_i=2, getfield_gc_r=2)
 
     def test_double_frame(self):
         myjitdriver = JitDriver(greens = [], reds = ['n', 'xy', 'other'],
@@ -230,8 +230,9 @@ class ExplicitVirtualizableTests:
             return xy.inst_x
         res = self.meta_interp(f, [20])
         assert res == 134
-        self.check_simple_loop(setfield_gc=1, getfield_gc=0, cond_call=1)
-        self.check_resops(setfield_gc=2, getfield_gc=4)
+        self.check_simple_loop(setfield_gc=1, getfield_gc_i=0, cond_call=1,
+                               getfield_gc_r=0)
+        self.check_resops(setfield_gc=2, getfield_gc_i=2, getfield_gc_r=2)
 
     # ------------------------------
 
@@ -282,7 +283,8 @@ class ExplicitVirtualizableTests:
         res = self.meta_interp(f, [16])
         assert res == 3001 + 16 * 80
         self.check_simple_loop(setarrayitem_gc=0, setfield_gc=0,
-                               getarrayitem_gc=0, getfield_gc=0)
+            getarrayitem_gc_i=0, getarrayitem_gc_r=0, getfield_gc_i=0,
+            getfield_gc_r=0)
 
     def test_synchronize_arrays_in_return(self):
         myjitdriver = JitDriver(greens = [], reds = ['n', 'xy2'],
@@ -313,8 +315,9 @@ class ExplicitVirtualizableTests:
         assert f(18) == 10360
         res = self.meta_interp(f, [18])
         assert res == 10360
-        self.check_simple_loop(setfield_gc=0, getarrayitem_gc=0,
-                               getfield_gc=0, setarrayitem_gc=0)
+        self.check_simple_loop(setfield_gc=0, getarrayitem_gc_i=0,
+                               getfield_gc_i=0, getfield_gc_r=0,
+                               setarrayitem_gc=0, getarrayitem_gc_r=0)
 
     def test_array_length(self):
         myjitdriver = JitDriver(greens = [], reds = ['n', 'xy2'],
@@ -340,8 +343,9 @@ class ExplicitVirtualizableTests:
             return xy2.inst_l1[1]
         res = self.meta_interp(f, [18])
         assert res == 2941309 + 18
-        self.check_simple_loop(setfield_gc=0, getarrayitem_gc=0,
-                               arraylen_gc=0, getfield_gc=0)
+        self.check_simple_loop(setfield_gc=0, getarrayitem_gc_i=0,
+                               getarrayitem_gc_r=0, getfield_gc_i=0,
+                               arraylen_gc=0, getfield_gc_r=0)
 
     def test_residual_function(self):
         myjitdriver = JitDriver(greens = [], reds = ['n', 'xy2'],
@@ -374,8 +378,9 @@ class ExplicitVirtualizableTests:
             return xy2.inst_l1[1]
         res = self.meta_interp(f, [18])
         assert res == 2941309 + 18
-        self.check_simple_loop(call=1, setfield_gc=0, getarrayitem_gc=0,
-                               arraylen_gc=1, getfield_gc=0)
+        self.check_simple_loop(call_r=1, setfield_gc=0, getarrayitem_gc_i=0,
+            getarrayitem_gc_r=0, arraylen_gc=1, getfield_gc_i=0,
+            getfield_gc_r=0)
 
     def test_double_frame_array(self):
         myjitdriver = JitDriver(greens = [], reds = ['n', 'xy2', 'other'],
@@ -412,7 +417,8 @@ class ExplicitVirtualizableTests:
         res = self.meta_interp(f, [20], enable_opts='')
         assert res == expected
         self.check_simple_loop(setarrayitem_gc=1, setfield_gc=0,
-                               getarrayitem_gc=1, arraylen_gc=1, getfield_gc=2)
+                               getarrayitem_gc_i=1, arraylen_gc=1,
+                               getfield_gc_r=2)
 
     # ------------------------------
 
@@ -461,7 +467,8 @@ class ExplicitVirtualizableTests:
         assert f(18) == 10360
         res = self.meta_interp(f, [18])
         assert res == 10360
-        self.check_simple_loop(getfield_gc=0, getarrayitem_gc=0,
+        self.check_simple_loop(getfield_gc_i=0, getfield_gc_r=0,
+                               getarrayitem_gc_i=0, getarrayitem_gc_r=0,
                                setfield_gc=0, setarrayitem_gc=0)
 
     # ------------------------------
@@ -495,7 +502,8 @@ class ImplicitVirtualizableTests(object):
 
         res = self.meta_interp(f, [10])
         assert res == 55
-        self.check_simple_loop(setfield_gc=0, getfield_gc=0)
+        self.check_simple_loop(setfield_gc=0, getfield_gc_i=0,
+                               getfield_gc_r=0)
 
     def test_virtualizable_with_array(self):
         myjitdriver = JitDriver(greens = [], reds = ['n', 'x', 'frame'],
@@ -529,8 +537,10 @@ class ImplicitVirtualizableTests(object):
 
         res = self.meta_interp(f, [10, 1], listops=True)
         assert res == f(10, 1)
-        self.check_simple_loop(getfield_gc=0, getarrayitem_gc=0)
-        self.check_resops(getfield_gc=2, getarrayitem_gc=4)
+        self.check_simple_loop(getfield_gc_i=0, getarrayitem_gc_i=0,
+                               getarrayitem_gc_r=0, getfield_gc_r=0)
+        self.check_resops(getfield_gc_r=1, getarrayitem_gc_i=4,
+                          getfield_gc_i=1)
 
     def test_subclass_of_virtualizable(self):
         myjitdriver = JitDriver(greens = [], reds = ['frame'],
@@ -558,7 +568,8 @@ class ImplicitVirtualizableTests(object):
 
         res = self.meta_interp(f, [10])
         assert res == 55
-        self.check_simple_loop(setfield_gc=0, getfield_gc=0)
+        self.check_simple_loop(setfield_gc=0, getfield_gc_i=0,
+                               getfield_gc_r=0)
 
     def test_external_pass(self):
         jitdriver = JitDriver(greens = [], reds = ['n', 'z', 'frame'],
@@ -1091,7 +1102,8 @@ class ImplicitVirtualizableTests(object):
 
         res = self.meta_interp(f, [10])
         assert res == 55
-        self.check_simple_loop(setfield_gc=0, getfield_gc=0)
+        self.check_simple_loop(setfield_gc=0, getfield_gc_i=0,
+                               getfield_gc_r=0)
 
         from rpython.jit.backend.test.support import BaseCompiledMixin
         if isinstance(self, BaseCompiledMixin):
@@ -1102,10 +1114,8 @@ class ImplicitVirtualizableTests(object):
                                        if getattr(graph, 'func', None) is f]
         init_graph = t._graphof(Frame.__init__.im_func)
 
-        deref = t.rtyper.type_system.deref
-
         def direct_calls(graph):
-            return [deref(op.args[0].value)._callable.func_name
+            return [op.args[0].value._obj._callable.func_name
                     for block, op in graph.iterblockops()
                         if op.opname == 'direct_call']
 
@@ -1251,8 +1261,10 @@ class ImplicitVirtualizableTests(object):
 
         res = self.meta_interp(f, [10])
         assert res == 155
-        self.check_simple_loop(setfield_gc=0, getfield_gc=0)
-        self.check_resops(setfield_gc=0, getfield_gc=2)
+        self.check_simple_loop(setfield_gc=0, getfield_gc_i=0,
+                               getfield_gc_r=0)
+        self.check_resops(setfield_gc=0, getfield_gc_i=2,
+                          getfield_gc_r=0)
 
     def test_blackhole_should_synchronize(self):
         myjitdriver = JitDriver(greens = [], reds = ['frame'],
@@ -1288,8 +1300,9 @@ class ImplicitVirtualizableTests(object):
 
         res = self.meta_interp(f, [10])
         assert res == 155
-        self.check_simple_loop(setfield_gc=0, getfield_gc=0)
-        self.check_resops(setfield_gc=0, getfield_gc=2)
+        self.check_simple_loop(setfield_gc=0, getfield_gc_i=0,
+                               getfield_gc_r=0)
+        self.check_resops(setfield_gc=0, getfield_gc_i=2, getfield_gc_r=0)
 
     def test_blackhole_should_not_reenter(self):
         if not self.basic:
@@ -1500,7 +1513,7 @@ class ImplicitVirtualizableTests(object):
         res = self.meta_interp(main, [10])
         assert res == main(10)
         self.check_resops({
-            "getfield_gc": 1, "int_lt": 2, "ptr_eq": 1, "guard_true": 3,
+            "getfield_gc_i": 1, "int_lt": 2, "ptr_eq": 1, "guard_true": 3,
             "int_add": 2, "jump": 1
         })
 
@@ -1606,7 +1619,7 @@ class ImplicitVirtualizableTests(object):
         if bridge is not None:
             l = [op for op in
                  bridge.operations if op.getopnum() == rop.SETFIELD_GC]
-            assert "'inst_x'" in str(l[0].getdescr().realdescrref())
+            assert "'inst_x'" in str(l[1].getdescr().realdescrref())
             assert len(l) == 2 # vable token set to null
             l = [op for op in bridge.operations if
                  op.getopnum() == rop.GUARD_NOT_FORCED_2]
