@@ -251,9 +251,7 @@ class ClassRepr(Repr):
         allmethods = {}
         # class attributes
         llfields = []
-        attrs = self.classdef.attrs.items()
-        attrs.sort()
-        for name, attrdef in attrs:
+        for name, attrdef in self.classdef.attrs.items():
             if attrdef.readonly:
                 s_value = attrdef.s_value
                 s_unboundmethod = self.prepare_method(s_value)
@@ -271,6 +269,8 @@ class ClassRepr(Repr):
             mangled_name = mangle('pbc%d' % counter, attr)
             pbcfields[access_set, attr] = mangled_name, r
             llfields.append((mangled_name, r.lowleveltype))
+        llfields.sort()
+        llfields.sort(key=attr_reverse_size)
         #
         self.rbase = getclassrepr(self.rtyper, self.classdef.basedef)
         self.rbase.setup()
@@ -492,19 +492,7 @@ class InstanceRepr(Repr):
                     fields[name] = mangled_name, r
                     myllfields.append((mangled_name, r.lowleveltype))
 
-            # Sort the instance attributes by decreasing "likely size",
-            # as reported by rffi.sizeof(), to minimize padding holes in C.
-            # Fields of the same size are sorted by name (by attrs.sort()
-            # above) just to minimize randomness.
-            def keysize((_, T)):
-                if T is lltype.Void:
-                    return None
-                from rpython.rtyper.lltypesystem.rffi import sizeof
-                try:
-                    return -sizeof(T)
-                except StandardError:
-                    return None
-            myllfields.sort(key=keysize)
+            myllfields.sort(key=attr_reverse_size)
             if llfields is None:
                 llfields = myllfields
             else:
@@ -696,10 +684,12 @@ class InstanceRepr(Repr):
             rbase = rbase.rbase
         return False
 
-    def new_instance(self, llops, classcallhop=None):
+    def new_instance(self, llops, classcallhop=None, nonmovable=False):
         """Build a new instance, without calling __init__."""
         flavor = self.gcflavor
         flags = {'flavor': flavor}
+        if nonmovable:
+            flags['nonmovable'] = True
         ctype = inputconst(Void, self.object_type)
         cflags = inputconst(Void, flags)
         vlist = [ctype, cflags]
@@ -1043,9 +1033,10 @@ class __extend__(pairtype(InstanceRepr, InstanceRepr)):
 
 # ____________________________________________________________
 
-def rtype_new_instance(rtyper, classdef, llops, classcallhop=None):
+def rtype_new_instance(rtyper, classdef, llops, classcallhop=None,
+                       nonmovable=False):
     rinstance = getinstancerepr(rtyper, classdef)
-    return rinstance.new_instance(llops, classcallhop)
+    return rinstance.new_instance(llops, classcallhop, nonmovable=nonmovable)
 
 def ll_inst_hash(ins):
     if not ins:
@@ -1072,6 +1063,19 @@ def fishllattr(inst, name, default=_missing):
                              (lltype.typeOf(widest), name))
     return default
 
+def attr_reverse_size((_, T)):
+    # This is used to sort the instance or class attributes by decreasing
+    # "likely size", as reported by rffi.sizeof(), to minimize padding
+    # holes in C.  Fields should first be sorted by name, just to minimize
+    # randomness, and then (stably) sorted by 'attr_reverse_size'.
+    if T is lltype.Void:
+        return None
+    from rpython.rtyper.lltypesystem.rffi import sizeof
+    try:
+        return -sizeof(T)
+    except StandardError:
+        return None
+
 
 # ____________________________________________________________
 #
